/*
  mainwindow.cpp
  Copyright (C) 2008-2009 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

//#define DEBUG

#include "mainwindow.h"

#include <qwidget.h>
#include <stdio.h>
#include "qgmap.h"
#include "qgconf.h"
#include "version.h"

#include <qdialog.h>
#include <qdatetime.h>

#include "pos.h"

#include "gmap.h"
#include "centerpos.h"
#include "slider.h"
#include "marker.h"
#include "iconmenu.h"
#include "scale.h"
#include "gpsview.h"
#include "poi_view.h"
#include "centermarker.h"
#include "freehandregion.h"
#include "poiinfoview.h"

#include "help.h"
#include "info.h"

#include "mounter.h"

#ifndef QWS
#include <qlabel.h>
#endif
#ifdef QWS
#include <qcopchannel_qws.h>
#endif

MainWindow::MainWindow( QWidget *parent, const char *name, WFlags fl) :
	LayeredBase(parent, name, fl),
	conf(this, "config")
{
#ifndef QWS
	resize(640,640);
#endif
	setCaption(QString::fromUtf8("qgmap - ")+QGMAP_VER);

	mounter = new Mounter(this, "mounter");

	gmap = new GMap(this,"gmap");
	center = new CenterPos(this,"center");
	slider = new Slider(this,"slider");
	marker = new Marker(this,"marker");
	menu = new IconMenu(this,"menu");
	scale = new Scale(this,"scale");
	gpsview = new GPSView(this,"gpsview");
	poiview = new PoiView(this,"poiview");
	centermarker = new CenterMarker(this,"centermarker");
	freehand = new FreehandRegion(this,"freehand");

	help = new Help(this);
	info = new Info(this);
	poiinfo = new PoiInfoView(this);

	/* config event */
	connect(&conf, SIGNAL(configChanged(const QString &, const QString &)),
			this, SIGNAL(configChanged(const QString &, const QString &)));

	addWidget(gmap);
	addWidget(center);
	addWidget(slider);
	addWidget(marker);
	addWidget(menu);
	addWidget(scale);
	addWidget(gpsview);
	addWidget(poiview);
	addWidget(centermarker);
	addWidget(freehand);

#ifndef QWS
	label = new QLabel(this);
	label->setGeometry(640-16,640-16,16,16);
#endif

#ifdef QWS
	QCopChannel *qcop = new QCopChannel(QCOPNAME, this);
	connect(qcop, SIGNAL(received(const QCString &, const QByteArray&)),
			this, SLOT(recvQCop(const QCString &, const QByteArray&)));
#endif
	conf.migrate();
}

MainWindow::~MainWindow()
{
	delete gmap;
	delete center;
	delete slider;
	delete marker;
	delete menu;
	delete scale;
	delete gpsview;
	delete centermarker;
	delete freehand;

	delete help;
	delete info;
	delete poiinfo;
}

void MainWindow::dispInfo(QString str)
{
	info->dispInfo(str);
}

void MainWindow::hideInfo()
{
	info->hide();
}

void MainWindow::dispPoiInfo(QPixmap const &icon, QString const &title,
							 QString const &desc, QString const &info)
{
	poiinfo->setIcon(icon);
	poiinfo->setTitle(title);
	poiinfo->setDesc(desc);
	poiinfo->setInfo(info);
	poiinfo->show();
}

void MainWindow::mousePressEvent( QMouseEvent *e )
{
	if(help->isVisible()){
		help->hide();
		return;
	}
	if(poiinfo->isVisible()){
		poiinfo->hide();
		return;
	}

	LayeredBase::mousePressEvent(e);
}

void MainWindow::keyPressEvent ( QKeyEvent * e )
{
	if(help->isVisible()){
		help->hide();
		return;
	}
	if(poiinfo->isVisible()){
		poiinfo->hide();
		return;
	}

	switch(e->key()){
	case Key_H:
#ifndef QWS // for PC
	case Key_Question:
#endif
		help->show();
		return;

	case Key_Period:
		if(e->state() & ShiftButton){
			help->show();
			return;
		}
		break;
	}

	LayeredBase::keyPressEvent(e);
}

void MainWindow::showEvent( QShowEvent *e )
{
	LayeredBase::showEvent(e);

	conf.reloadEntry("MapPath0", DEFAULT_MAP_PATH);
	conf.reloadEntry("MapPath1", "");
	conf.reloadEntry("MapPath2", "");
	conf.reloadEntry("MapPath3", "");
	conf.reloadEntry("MapPath4", "");
	conf.reloadEntry("MapPath5", "");
	conf.reloadEntry("MapPath6", "");
	conf.reloadEntry("MapPath7", "");

	conf.reloadEntry("Gamma", DEFAULT_GAMMA);

	conf.reloadEntry("GPSDev", DEFAULT_GPS_DEVICE);
	conf.reloadEntry("PreCommand", DEFAULT_PRE_COMMAND);
	conf.reloadEntry("PostCommand", DEFAULT_POST_COMMAND);
	conf.reloadEntry("GPSInterval", STR(DEFAULT_GPS_INTERVAL));
	conf.reloadEntry("GPSDatum", STR((int)DEFAULT_GPS_DATUM));
	conf.reloadEntry("GPSTimeZone", STR(DEFAULT_GPS_TIMEZONE));

	center->loadPosition(0);
}

void MainWindow::paintEvent( QPaintEvent *e )
{
	LayeredBase::paintEvent(e);
	if(info->isVisible())
		info->repaint();
	if(help->isVisible())
		help->repaint();
	if(poiinfo->isVisible())
		poiinfo->repaint();
}

void MainWindow::recvQCop(const QCString &m, const QByteArray &d)
{
	if(m == "setWGSLocation(QString)"){
		QString str;
		QDataStream in(d, IO_ReadOnly);
		in >> str;

		Pos pos(center->getPosition());
		pos.setPos(str, true);

		center->setPosition(pos);
	} else if(m == "setTokyoLocation(QString)"){
		QString str;
		QDataStream in(d, IO_ReadOnly);
		in >> str;

		Pos pos(center->getPosition());
		pos.setPos(str, false);

		center->setPosition(pos);
	}
}
