/*
  poi_db.cpp
  Copyright (C) 2009 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

#include "qgmap.h"
#include "sqlite.h"
#include "poi_db.h"
#include <assert.h>
#include <qfile.h>
#include <qtextcodec.h>
#include <qtextstream.h>
#include <qpixmap.h>

#include "pos.h"

PoiDb::PoiDb(const char *db_file)
{
	db = new SQLite();
	if (db->open(db_file)) {
		delete db;
		db = NULL;
	}

	db->exec("CREATE TABLE IF NOT EXISTS file ( "					\
			 "id INTEGER PRIMARY KEY AUTOINCREMENT, "				\
			 "grp_id INTEGER(2) NOT NULL, "							\
			 "icon_id INTEGER(2) NOT NULL, "						\
			 "enable INTEGER(1) DEFAULT 1, "						\
			 "zoom INTEGER(1) DEFAULT 7, "							\
			 "desc TEXT DEFAULT ''"									\
			 ")");

	db->exec("CREATE TABLE IF NOT EXISTS icon ( "					\
			 "id INTEGER PRIMARY KEY AUTOINCREMENT, "				\
			 "width INTEGER(1) NOT NULL, "							\
			 "height INTEGER(1) NOT NULL, "							\
			 "icon_data BLOB NOT NULL"								\
			 ")");

	db->exec("CREATE TABLE IF NOT EXISTS poi ( "					\
			 "id INTEGER PRIMARY KEY AUTOINCREMENT, "				\
			 "file_id INTEGER(2) NOT NULL, "						\
			 "lat INTEGER(4) NOT NULL, "							\
			 "lon INTEGER(4) NOT NULL, "							\
			 "name TEXT DEFAULT '', "								\
			 "info TEXT DEFAULT ''"									\
			 ")");

	db->exec("CREATE INDEX IF NOT EXISTS idx_poi ON poi(lat, lon)");

	db->exec("CREATE TABLE IF NOT EXISTS grp ( "					\
			 "id INTEGER PRIMARY KEY AUTOINCREMENT, "				\
			 "name TEXT NOT NULL"									\
			 ")");

	db->exec("PRAGMA cache_size = 512;");  /* 512kB cache */
}

PoiDb::~PoiDb()
{
	if(db) {
		delete db;
		db = NULL;
	}
}


PoiDbItem *PoiDb::queryPoiMapItem(int x, int y, int zoom, int limit, bool *too_many)
{
	assert(db);
	int size;
	int mx, my;
	int icon_size = (16 - 1) << zoom;

	if (too_many)
		*too_many = false;

	x &= ~255;
	y &= ~255;

	mx = x << zoom;
	my = y << zoom;
	size = 256 << zoom;

	SQLiteStatement *select = db->newQuery(
		"SELECT COUNT(*) FROM poi as p, file as f "						\
		"WHERE f.enable = 1 AND f.zoom >= ? AND "						\
		"p.lat BETWEEN ? AND ? AND p.lon BETWEEN ? AND ? AND "			\
		"p.file_id = f.id");
	if (!select)
		return NULL;

	select->bind(1, zoom);
	select->bind(2, mx - icon_size);
	select->bind(3, mx + icon_size + size - 1);
	select->bind(4, my - icon_size);
	select->bind(5, my + icon_size + size - 1);

	if (select->step() != SQLITE_ROW) {
		delete select;
		return NULL;
	}

	if (select->getColumnInt(0) > limit) {
		delete select;
		if (too_many)
			*too_many = true;
		return NULL;
	}

	delete select;

	select = db->newQuery(
		"SELECT p.id, f.icon_id, p.lat, p.lon FROM poi as p, file as f " \
		"WHERE f.enable = 1 AND f.zoom >= ? AND "						\
		"p.lat BETWEEN ? AND ? AND p.lon BETWEEN ? AND ? AND "			\
		"p.file_id = f.id");
	if (!select)
		return NULL;

	select->bind(1, zoom);
	select->bind(2, mx - icon_size);
	select->bind(3, mx + icon_size + size - 1);
	select->bind(4, my - icon_size);
	select->bind(5, my + icon_size + size - 1);

//	printf("x:%d-%d y:%d-%d\n", mx, mx + size - 1, my, my + size - 1);

	PoiDbItem *item = new PoiDbItem(select, x, y, zoom);

	if (!item)
		delete select;

	return item;
}

PoiDbItem *PoiDb::queryPoiItem(int x, int y, int zoom)
{
	assert(db);
	int mx, my;
	int icon_size = ((16 - 1) << zoom) / 2;

	mx = x << zoom;
	my = y << zoom;

	SQLiteStatement *select = db->newQuery(
		"SELECT p.id, f.icon_id, p.lat, p.lon, p.name, p.info, p.file_id FROM poi as p, file as f " \
		"WHERE f.enable = 1 AND f.zoom >= ? AND "						\
		"p.lat BETWEEN ? AND ? AND p.lon BETWEEN ? AND ? AND "			\
		"p.file_id = f.id");
	if (!select)
		return NULL;

	select->bind(1, zoom);
	select->bind(2, mx - icon_size);
	select->bind(3, mx + icon_size - 1);
	select->bind(4, my - icon_size);
	select->bind(5, my + icon_size - 1);

	PoiDbItem *item = new PoiDbItem(select, 0, 0, 0);

	if (!item)
		delete select;

	return item;
}

PoiDbFile *PoiDb::queryPoiFile(int grp_id, int id)
{
	assert(db);

	QString q;

	if (grp_id < 0) {
		if (id < 0)
			return NULL;
		q = QString("SELECT * FROM file WHERE id = %1").arg(id);
	} else {
		if (id < 0)
			q = QString("SELECT * FROM file WHERE grp_id = %1").arg(grp_id);
		else
			q = QString("SELECT * FROM file WHERE grp_id = %1 AND id = %2").arg(grp_id).arg(id);
	}

	SQLiteStatement *select = db->newQuery(q);
	if (!select)
		return NULL;

	select->bind(1, grp_id);

	PoiDbFile *item = new PoiDbFile(select);

	if (!item)
		delete select;

	return item;
}

PoiDbIcon *PoiDb::queryPoiIcon()
{
	assert(db);

	SQLiteStatement *select = db->newQuery("SELECT * FROM icon");
	if (!select)
		return NULL;

	PoiDbIcon *icon = new PoiDbIcon(select);

	if (!icon)
		delete select;

	return icon;
}

int PoiDb::getNewFileId()
{
	int ret = db->execOneColumn("SELECT MAX(id) FROM file id");

	if(ret < 0)
		return 1;

	return ret + 1;
}

int PoiDb::getNewIconId()
{
	int ret = db->execOneColumn("SELECT MAX(id) FROM icon id");

	if(ret < 0)
		return 1;

	return ret + 1;
}

int PoiDb::importPoiCSV(QString file_path, int icon_id, int zoom, QString desc)
{
	assert(db);

	QFile file(file_path);
	QString q;

	if (!file.open(IO_ReadOnly))
		return -1;

	QTextStream t(&file);
	t.setCodec(QTextCodec::codecForName("ShiftJIS"));

	q = "INSERT INTO poi (file_id, lat, lon, name, info) VALUES (?, ?, ?, ?, ?)";

	SQLiteStatement *insert = db->newQuery(q);
	if (!insert) {
		file.close();
		return -1;
	}

	int fid = getNewFileId();
	q = "INSERT INTO file (id, grp_id, icon_id, zoom, desc) VALUES (%1, 0, %2, %3, '%4')";
	int ret = db->exec(q.arg(fid).arg(icon_id).arg(zoom).arg(SQLite::escape(desc)).utf8());
	if (ret) {
		file.close();
		delete insert;
		return -1;
	}

	db->exec("BEGIN");

	while (!t.eof()) {
		QString str = t.readLine();
		while (str.contains('"') % 2 == 1 && !t.eof()) {
			str += "\n" + t.readLine();
		}

		QStringList list = splitCSV(str);

		if(list.count() < 2 || list.count() > 4)
			continue;

		Pos pos(list[1].toDouble(), list[0].toDouble(), 0);
//		printf("lat,lon (%f,%f) -> x,y (%d,%d)\n", list[1].toDouble(), list[0].toDouble(),
//			   pos.getX(), pos.getY());

		insert->reset();
		insert->bind(1, fid);
		insert->bind(2, pos.getX());
		insert->bind(3, pos.getY());
		if(list.count() >= 3)
			insert->bind(4, list[2]);
		else
			insert->bindNull(4);
		if(list.count() >= 4)
			insert->bind(5, list[3]);
		else
			insert->bindNull(5);

		ret = insert->step();
	}

	delete insert;
	file.close();

	db->exec("COMMIT");

	return fid;
}

int PoiDb::addIcon(const QByteArray &data)
{
	assert(db);

	QString q = "INSERT INTO icon (id, width, height, icon_data) VALUES (?, ?, ?, ?)";

	SQLiteStatement *insert = db->newQuery(q);
	if (!insert)
		return -1;

	int id = getNewIconId();

	QPixmap icon(data);

	insert->bind(1, id);
	insert->bind(2, icon.width());
	insert->bind(3, icon.height());
	insert->bind(4, data);

	int ret = insert->step();
//	printf("ret = %d\n", ret);

	delete insert;

	return id;
}

int PoiDb::delIcon(int id)
{
	assert(db);

	QString q = "UPDATE file SET icon_id = 0 WHERE icon_id = %1;"
		"DELETE FROM icon WHERE id = %2;";

	int ret = db->exec(q.arg(id).arg(id));
//	printf("ret =%d\n",ret);
	return ret;
}

int PoiDb::delFile(int id)
{
	QString q;
	int ret;
	assert(db);

	db->exec("BEGIN");

	q = "DELETE FROM file WHERE id = %1;";
	ret = db->exec(q.arg(id));

	q = "DELETE FROM poi WHERE file_id = %1;";
	ret = db->exec(q.arg(id));

	db->exec("COMMIT");

	return 0;
}

QByteArray PoiDb::getPoiIcon(int id)
{
	assert(db);

	SQLiteStatement *select = db->newQuery("SELECT icon_data FROM icon WHERE id = ?");
	if (!select)
		return NULL;

	select->bind(1, id);

	int ret = select->step();
//	printf("ret = %d\n", ret);

	if (ret != SQLITE_ROW)
		return QByteArray();

	QByteArray data = select->getColumnBlob(0);
	delete select;

	return data;
}

int PoiDb::changePoiFile(int id, int icon_id, int zoom, QString name)
{
	assert(db);

	SQLiteStatement *update = db->newQuery("UPDATE file SET icon_id = ?, zoom = ?, desc = ? WHERE id = ?");
	if (!update)
		return 0;

	update->bind(1, icon_id);
	update->bind(2, zoom);
	update->bind(3, name);
	update->bind(4, id);

	int ret = update->step();
//	printf("ret = %d\n", ret);

	if (ret != SQLITE_OK)
		return 0;

	return 1;
}

int PoiDb::changePoiFileEnable(int id, int enable)
{
	assert(db);

	SQLiteStatement *update = db->newQuery("UPDATE file SET enable = ? WHERE id = ?");
	if (!update)
		return 0;

	update->bind(1, enable);
	update->bind(2, id);

	int ret = update->step();
//	printf("ret = %d\n", ret);

	if (ret != SQLITE_OK)
		return 0;

	return 1;
}

QStringList PoiDb::splitCSV(QString str)
{
	QStringList list;
	QString tmp;
	int i;
	bool instr = false;

	str = str.stripWhiteSpace();

	for (i = 0; i < (int)str.length(); i++) {
		QChar c = str[i];

		if (!instr) {
			switch (c.unicode()) {
			case '"':
				instr = true;
				continue;
			case ',':
				list.append(tmp.stripWhiteSpace());
				tmp = "";
				continue;
			}
		} else {
			switch (c.unicode()) {
			case '"':
				if (str[i+1].unicode() == '"') {
					tmp.append(c);
					i++;
				} else {
					instr = false;
				}
				continue;
			case '\\':
				tmp.append(str[i+1]);
				i++;
				continue;
			}
		}
		tmp.append(c);
	}

	tmp = tmp.stripWhiteSpace();
	if (tmp.length() > 0)
		list.append(tmp);

	return list;
}

/************ PoiDbResult ************/

PoiDbResult::PoiDbResult(SQLiteStatement *s)
{
	stmt = s;
}

PoiDbResult::~PoiDbResult()
{
	delete stmt;
}

bool PoiDbResult::is_end()
{
	return status != SQLITE_ROW;
}

bool PoiDbResult::next()
{
	status = stmt->step();
	return status == SQLITE_ROW;
}

int PoiDbResult::count()
{
	return stmt->getRowCount();
}

/************ PoiDbItem ************/

PoiDbItem::PoiDbItem(SQLiteStatement *s, int x, int y, int zoom) :
	PoiDbResult(s)
{
	map_x = x;
	map_y = y;
	map_zoom = zoom;
}

int PoiDbItem::getId()
{
	return stmt->getColumnInt(0);
}

int PoiDbItem::getIconId()
{
	return stmt->getColumnInt(1);
}

int PoiDbItem::getX()
{
	return (stmt->getColumnInt(2) >> map_zoom) - map_x;
}

int PoiDbItem::getY()
{
	return (stmt->getColumnInt(3) >> map_zoom) - map_y;
}

QString PoiDbItem::getName()
{
	return stmt->getColumnText(4);
}

QString PoiDbItem::getInfo()
{
	return stmt->getColumnText(5);
}

int PoiDbItem::getFileId()
{
	return stmt->getColumnInt(6);
}

/************ PoiDbFile ************/

PoiDbFile::PoiDbFile(SQLiteStatement *s) :
	PoiDbResult(s)
{
}

int PoiDbFile::getId()
{
	return stmt->getColumnInt(0);
}

int PoiDbFile::getGrpId()
{
	return stmt->getColumnInt(1);
}

int PoiDbFile::getIconId()
{
	return stmt->getColumnInt(2);
}

int PoiDbFile::getEnabled()
{
	return stmt->getColumnInt(3);
}

int PoiDbFile::getZoom()
{
	return stmt->getColumnInt(4);
}

QString PoiDbFile::getName()
{
	return stmt->getColumnText(5);
}

/************ PoiDbIcon ************/

PoiDbIcon::PoiDbIcon(SQLiteStatement *s) :
	PoiDbResult(s)
{}

int PoiDbIcon::getId()
{
	return stmt->getColumnInt(0);
}

int PoiDbIcon::getWidth()
{
	return stmt->getColumnInt(1);
}

int PoiDbIcon::getHeight()
{
	return stmt->getColumnInt(2);
}

QByteArray PoiDbIcon::getIconData()
{
	return stmt->getColumnBlob(3);
}
